# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import os


def change_mod(mode, path, flags=""):
    if path is None:
        path = "."
    if os.path.isdir(path):
        flags = " -R"
    return "chmod%s %s %s" % (flags, mode, path)


def change_own(owner, group, path, flags=""):
    if path is None:
        path = "."
    if os.path.isdir(path):
        flags = " -R"
    if owner is not None and group is not None:
        return "chown%s %s:%s %s" % (flags, owner, group, path)
    elif owner is None and group is not None:
        return "chgrp%s %s %s" % (flags, group, path)
    elif owner is not None and group is None:
        return "chown%s %s %s" % (flags, owner, path)
    return "echo cannot chown"


def copy_files(src, dst, flags=""):
    if src is None and dst is None:
        return "echo invalid copy command"
    elif src is None:
        src = "."
    elif dst is None:
        dst = "."
    if os.path.isdir(src):
        flags = " -R"
    return "cp%s %s %s" % (flags, src, dst)


def move_files(src, dst):
    if src is None and dst is None:
        return "echo invalid move command"
    elif src is None:
        src = "."
    elif dst is None:
        dst = "."
    return "mv %s %s" % (src, dst)


def remove_files(src, flags=""):
    if src is None:
        src = "."
    if os.path.isdir(src):
        flags = " -r"
    if flags == "":
        flags = " -f"
    else:
        flags += "f"
    return "rm%s %s 2> /dev/null" % (flags, src)


def make_symlink(src, dst):
    if src is None and dst is None:
        return "echo invalid ln command"
    elif src is None:
        src = "."
    elif dst is None:
        dst = "."
    return "ln -s %s %s" % (src, dst)


def unpack_bash():
    lang = {
        "exec_web": lambda url: "echo Cannot open URL in browser: %s" % url,
        "exec_shell": lambda cmd: cmd,
        "change_dir": lambda path: "cd %s" % path,
        "change_mod": change_mod,
        "change_own": change_own,
        "copy_files": copy_files,
        "move_files": move_files,
        "remove_files": remove_files,
        "make_symlink": make_symlink,
        "make_dir": lambda path: "mkdir -p %s" % path,
        "kill_proc": lambda pid, pname: "kill %s" % pid,
    }
    return lang, "bash.sh", """#!/bin/bash

{}
"""



# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from os import path
from glob import glob
from shutil import copy
from distutils.dir_util import copy_tree

def copy_files(src=None, dst=None, *args, **kwargs):
    r"""Copy files from a given source to a given destination.

    Args:
        src (str): Source of files.
        dst (str): Target destination of files.

    Retuns:
        str: Status of copy.

    Raises:
        OSError: If an invalid `src` or `dst` is provided.

    Accepted statements:
        ^copy from `(?P<src>.+)` to `(?P<dst>.+)`[\.\?\!]$
        ^copy `(?P<src>.+)` files to `(?P<dst>.+)`[\.\?\!]$
        ^copy `(?P<src>.+)` to `(?P<dst>.+)`[\.\?\!]$

    Sample input:
        1) Run `touch /tmp/buildok_test_copy.txt`.
        2) Copy `/tmp/buildok_test_copy.txt` to `/tmp/buildok_test_copy2.txt`.

    Expected:
        Copied 1 file(s) and 0 folder(s)
    """
    files, folders = 0, 0
    for item in glob(src):
        if path.isfile(item):
            copy(item, dst)
            files += 1
        elif path.isdir(item):
            copy_tree(item, dst)
            folders += 1
    return "Copied %d file(s) and %d folder(s)" % (files, folders)



# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from pwd import getpwnam
from grp import getgrnam
from os import chown, getcwd

def change_own(owner="", group="", path=None, *args, **kwargs):
    r"""Change owner and group on file or directory.

    Args:
        owner (str): User name.
        group (str): Group name.
        path (str): Path to file or directory.

    Retuns:
        str: Human readable descriptor message or error.

    Raises:
        OSError: If an invalid `path` is provided.

    Accepted statements:
        ^change file owner to `(?P<owner>.+)` on `(?P<path>.+)`[\.\?\!]$
        ^change owner to `(?P<owner>.+)` on `(?P<path>.+)`[\.\?\!]$
        ^change user to `(?P<owner>.+)` on `(?P<path>.+)`[\.\?\!]$
        ^change user and group to `(?P<owner>.+):(?P<group>.+)`[\.\?\!]$
        ^set owner and group `(?P<owner>.+):(?P<group>.+)` for `(?P<path>.+)`[\.\?\!]$

    Sample (input):
        1) Run `touch /tmp/buildok_test.txt`.
        2) Change owner to `nobody` on `/tmp/buildok_test.txt`.

    Expected:
        Changed owner nobody => /tmp/buildok_test.txt
    """
    try:
        if owner != "":
            uid = getpwnam(owner).pw_uid
        else:
            uid = -1
        if group != "":
            gid = getgrnam(group).gr_gid
        else:
            gid = -1
        if path is None:
            path = getcwd()
        chown(path, uid, gid)
        if uid != -1 and gid == -1:
            return "Changed owner %s => %s" % (owner, path)
        elif uid == -1 and gid != -1:
            return "Changed group %s => %s" % (group, path)
        elif uid != -1 and gid != -1:
            return "Changed ownwer:group %s:%s => %s" % (owner, group, path)
    except OSError as e:
        raise e
    return "Nothing to do"



# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from buildok.parser import Parser

def analyze(Stmt):
    """Analyze all statements.

    Returns:
        list: List of all statements including duplicated.
    """
    results, stmts = [], set([])
    for idx1, func in enumerate(Stmt.known_actions):
        class_ = func.__name__
        lines = func.__doc__.split("\n")
        funcs = Parser.parse(lines, [Stmt.statement_header], False)
        for idx2, stmt in enumerate(funcs):
            status = "duplicated" if stmt in stmts else "ok"
            line = ("%d.%d" % (idx1+1, idx2+1), class_, stmt, status)
            results.append(line)
            stmts.add(stmt)
    set_max = lambda o, s: len(s) if len(s) > o else o
    ids_max_len, grp_max_len, stmt_max_len, status_max_len = 0, 0, 0, 0
    for ids, grp, stmt, status in results:
        ids_max_len, grp_max_len = set_max(ids_max_len, ids), set_max(grp_max_len, grp)
        stmt_max_len, status_max_len = set_max(stmt_max_len, stmt), set_max(status_max_len, status)
    line = "| %-" + str(ids_max_len) + "s | %-" + str(grp_max_len) + "s | "
    line += "%-" + str(stmt_max_len) + "s | %-" + str(status_max_len) + "s |"
    header = line % ("", "Group", "Statement", "")
    sep = "-" * len(header)
    lines = [header, sep]
    for r in results:
        color = "\033[101m" if r[-1] == "duplicated" else "\033[94m"
        lines.append((color + line + "\033[0m") % r)
    lines.append(sep)
    return lines



# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from buildok.statement import Statement
from buildok.parser import Parser
from buildok.reader import Reader

from buildok.readers.file import FileReader
from buildok.readers.read_me import ReadmeReader

from buildok.util.console import Console, timeit_log
from buildok.util.shell import Shell
from buildok.util.analyze import analyze
from buildok.util.locker import lock, unlock

from converter import Converter


def setup():
    """Setup project path and verbose level.

    Raises:
        SystemExit: If any error occurs while parsing command line arguments.

    Return:
        bool: True if setup is done, otherwise False.
    """
    args = Shell.parse()
    Console.verbose = args.verbose
    if args.analyze:
        [Console.log(s) for s in analyze(Statement)]
        return None
    if args.project is not None:
        Reader.filepath = args.project
    if args.convert is not None:
        Converter.prepare(args.convert, Statement)
    return True


def read(first=True):
    """Read all posible sources.

    Args:
        first (bool): Return steps from first source only.

    Return:
        tuple: Returns build steps.
    """
    fr = FileReader()
    if not fr.exists():
        rr = ReadmeReader()
        if not rr.exists():
            raise Console.fatal("Nothing to build from...")
        else:
            Console.info("Building from README")
    else:
        Console.info("Building from file")
    steps = Reader.get_first() if first else Reader.get_all()
    for s in steps:
        if Parser.is_valid(s):
            Console.info("  %s" % s)
        else:
            Console.warn("  %s <--- bad grammar" % s)
    return steps


def run(steps, last_step="n/a"):
    """Parse steps and build.

    Args:
        steps (tuple): Ordonated list of build steps.
        last_step (str): Last known step.

    Raises:
        SystemExit: If build steps are invalid.
        IOError: If build steps link to invalid I/O operations.
    """
    pr = Parser(steps)
    pr.prepare(validate=True)
    while pr.has_step():
        step = pr.get_step()
        if step is None:
            raise Console.fatal("Unexpected step found after: %s" % last_step)
        Console.info(step)
        stmt = Statement(step)
        results = stmt.run()
        Console.eval(results)
        last_step = step
    Converter.check() and Converter.save()


@timeit_log
def main():
    lock()
    if not setup():
        return unlock()
    steps = read()
    if len(steps) == 0:
        unlock()
        raise Console.fatal("Nothing to build")
    try:
        run(steps)
    except Exception as e:
        unlock()
        raise Console.fatal(e)
    return unlock()



# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from buildok.reader import Reader

class FileReader(Reader):
    """File class helper to detect, create and read a certain given file.

    Attributes:
        filename (string): Path to build filename. Default is ".build".
        sample_build (string): Sample content of a build file.
    """
    READ_MODE, WRITE_MODE = "rb", "wb"
    filename = r".build"
    sample_build = r"1) Open browser `https://github.com/lexndru/buildok`!"

    def read(self):
        """Read build file content.

        Returns:
            str: Returns `filename` content on success.
            NoneType: Returns None on failure.

        Raises:
            IOError: If `filename` does not exist or does not have permissions.
        """
        with open(self.filename, self.READ_MODE) as file_:
            return [l.rstrip() for l in file_.readlines()]
        return None

    def create(self):
        """Write a sample build file with a dummy content.

        Returns:
            bool: Returns True on successful write, otherwise False.

        Raises:
            IOError: If permissions over `filename` are not enough or disk is full.
        """
        with open(self.filename, self.WRITE_MODE) as file_:
            file_.write(self.sample_build)
            return True
        return False



# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from os import kill
from signal import SIGTERM
from subprocess import check_output, CalledProcessError

def kill_proc(pid=None, pname=None, *args, **kwargs):
    r"""Send SIGTERM signal to a process.

    Args:
        pid (int): Process ID.
        pname (str): Process name.

    Retuns:
        str: Human readable descriptor message or error.

    Raises:
        OSError: If an invalid `pid` or `pname` is provided.
        CalledProcessError: If `pname` is not found.

    Accepted statements:
        ^kill process `(?P<pname>.+)`[\.\?\!]$
        ^kill pid `(?P<pid>.+)`[\.\?\!]$
        ^stop process `(?P<pname>.+)`[\.\?\!]$
        ^stop pid `(?P<pid>.+)`[\.\?\!]$
        ^nothing to do[\.\?\!]$

    Sample (input):
        1) Stop process `someProcessName`.

    Expected:
        Terminated process PID => 9999
    """
    try:
        if pname is not None:
            pid = check_output(["pidof", "-s", name])
        if pid is None:
            raise ValueError
        kill(int(pid), SIGTERM)
        return "Terminated process PID => %s" % pid
    except OSError as e:
        raise e
    except CalledProcessError as e:
        raise e
    except ValueError:
        return "Nothing to do"
    return "Nothing to do"



# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import webbrowser as wb

def exec_web(url=None, *args, **kwargs):
    r"""Open a link in default browser.

    Args:
        url (str): URL to open.

    Retuns:
        str: Output as string.

    Raises:
        TypeError: If an invalid `url` is provided.

    Accepted statements:
        ^open in browser `(?P<url>.+)`[\.\?\!]$
        ^open link `(?P<url>.+)`[\.\?\!]$
        ^open url `(?P<url>.+)`[\.\?\!]$

    Sample (input):
        1) Open link `https://github.com/lexndru/buildok`.

    Expected:
        Opened URL in browser => https://github.com/lexndru/buildok
    """
    try:
        wb.get().open(url, new=2)
        return "Opened URL in browser => %s" % url
    except TypeError as e:
        raise e
    return "Nothing to do"



# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from os import makedirs

def make_dir(path=None, *args, **kwargs):
    r"""Make a directory or make recursive directories.

    Args:
        path (str): Path to directory.

    Retuns:
        str: Human readable descriptor message or error.

    Raises:
        OSError: If an invalid `path` is provided or if path already exists.

    Accepted statements:
        ^create folder `(?P<path>.+)`[\.\?\!]$
        ^create directory `(?P<path>.+)`[\.\?\!]$
        ^make new folder `(?P<path>.+)`[\.\?\!]$
        ^make new directory `(?P<path>.+)`[\.\?\!]$

    Sample input:
        1) Go to `/tmp`.
        2) Create folder `buildok_test_folder`.

    Expected:
        Created new directory => buildok_test_folder
    """
    try:
        makedirs(path)
        return "Created new directory => %s" % path
    except OSError as e:
        raise e
    return "Nothing to do"



# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import re

class Parser(object):
    """Build steps parser.

    Attributes:
        steps (tuple): Ordonated list of build steps.
        build (list): List of ready-to-run instructions.

    Raises:
        TypeError: If provided `steps` are not a tuple.
    """
    PATTERN = re.compile(r"^\d+(?:\.\d+)?\)\s*(?P<step>.+)[^\!\.\?][\!\.\?]{1}$", re.U)

    def __init__(self, steps):
        if not isinstance(steps, tuple):
            raise TypeError("Unexpected steps datatype")
        self.steps, self.build = steps, [None] * len(steps)

    def prepare(self, validate=False):
        """Prepare, parse and sort build steps.

        Args:
            validate (bool): Stop on first error if True.

        Returns:
            self: Parser
        """
        for s in self.steps:
            if not Parser.is_valid(s):
                self.build.pop()
                continue
            sep = s.find(")")
            pos = int(s[:sep])
            self.build[pos-1] = s[sep+1:].lstrip()
        self.build.reverse()
        return self

    def has_step(self):
        """Check if there is a next step available.

        Returns:
            bool: True if steps are in queue, otherwise False.
        """
        return len(self.build) > 0

    def get_step(self):
        """Return step as string.

        Raises:
            IndexError: If no more steps are in queue.

        Returns:
            str: Raw step string.
        """
        return self.build.pop()

    @classmethod
    def is_valid(cls, step):
        """Test if a step is valid.

        Args:
            step (str): Raw step to test.

        Returns:
            bool: True if `step` has a valid pattern, otherwise False.
        """
        result = cls.PATTERN.match(step)
        try:
            step = result.group("step")
            return len(step) > 1
        except:
            return False

    @classmethod
    def all_valid(cls, steps):
        """Test if all steps are valid.

        Args:
            steps (list): List of steps to test.

        Returns:
            bool: True if all steps are valid, otherwise False.
        """
        if len(steps) == 0:
            return False
        return all([cls.is_valid(s) for s in steps])

    @classmethod
    def parse(cls, ctx=[], headlines=[], test=False):
        """Parse ctx and look for build steps.

        Args:
            ctx (list): Original content as lines.

        Returns:
            list: Non-empty list if build steps are found.
        """
        steps_start, steps_stop = -1, -1
        steps, newlines = [], []
        for idx, line_ in enumerate(ctx):
            line = line_.rstrip()
            if steps_start > -1:
                if len(newlines) > 0 and newlines[-1] > steps_start + 1:
                    break
                steps.append(line)
            if len(line) == 0:
                newlines.append(idx)
            for section in headlines:
                if section in line.lower():
                    if test and not cls.test_steps(ctx, idx):
                        continue
                    steps_start = idx
                    break
            if len(newlines) > 2:
                a, b = newlines[:-3:-1]
                if a == b + 1:
                    steps_stop = idx
                    break
        return [s.strip() for s in steps if s]

    @classmethod
    def test_steps(cls, ctx, start_index):
        """Test if a section contains valid steps.

        Args:
            ctx (list): Steps context.

        Returns:
            bool: True if section contains valid steps, otherwise False.
        """
        for line in ctx[start_index+1:]:
            if line != "":
                return cls.is_valid(line)
        return False



# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import os

class Reader(object):
    """Reader wrapper to gather build steps.

    A new instance registers a new reader.
    Has support for `.build` files and `README.md` files.

    Attributes:
        readers (list): A list of all registered readers.
        reader (str): Reader class name.
        filename (str): File name of the file to read.
        filepath (str): File path of the file to read.
    """
    readers = []
    reader = ""
    filename = ""
    filepath = ""

    def __init__(self):
        self.readers.append(self)
        self.reader = self.__class__.__name__
        self.patch_filename()

    def patch_filename(self):
        """Apply a patch to filename if filepath is set.

        Returns:
            self: Reader instance.
        """
        if self.filepath == "":
            return self
        if os.path.isfile(self.filepath):
            self.filename = self.filepath
        elif os.path.isdir(self.filepath):
            self.filename = r"{}/{}".format(self.filepath, self.filename)
        return self

    def read(self):
        """Read build file content.

        Raises:
            NotImplementedError: Readers must implement this method.
        """
        raise NotImplementedError("%s must implement read method" % self.reader)

    def exists(self):
        """Check if build file exists.

        Returns:
            bool: Returns True if `filename` exists, otherwise False.
        """
        return os.path.isfile(self.filename)

    @classmethod
    def get_all(cls):
        """Return all build steps from all sources.

        Raises:
            ValueError: If no reader was initialized.

        Returns:
            tuple: Non-empty tuple of build steps.
        """
        if len(cls.readers) == 0:
            raise ValueError("No reader initialized")
        data = []
        for r in cls.readers:
            data += r.read()
        return tuple(data)

    @classmethod
    def get_first(cls):
        """Return all build steps from first source.

        Raises:
            ValueError: If no reader was initialized.

        Returns:
            tuple: Non-empty tuple of build steps.
        """
        try:
            data = cls.readers[0].read()
        except IndexError:
            raise ValueError("No reader initialized")
        return tuple([d for d in data if d])



# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import re

from buildok.parser import Parser

from buildok.statements.web import exec_web
from buildok.statements.shell import exec_shell
from buildok.statements.chdir import change_dir
from buildok.statements.chmod import change_mod
from buildok.statements.chown import change_own
from buildok.statements.mkdir import make_dir
from buildok.statements.kill import kill_proc
from buildok.statements.copy import copy_files
from buildok.statements.move import move_files
from buildok.statements.remove import remove_files
from buildok.statements.symlink import make_symlink

class Statement(object):
    """Statement parser.

    Attributes:
        statement_header (str): Lookup string before build steps.
        known_actions (frozen set): Set of all known statements and actions.
        step (str): Raw step statement.
        stmt (func): Matched statement from `known_actions`.
        statements (list): List of all statements.
    """
    statements = []
    statement_header = r"accepted statements"
    known_actions = {
        exec_web,
        exec_shell,
        change_dir,
        change_mod,
        change_own,
        copy_files,
        move_files,
        remove_files,
        make_symlink,
        make_dir,
        kill_proc,
    }

    def __init__(self, step):
        self.step = step
        self.stmt, self.args = self.parse()
        self.statements.append(self)

    def __str__(self):
        return "Step: %s\nStatement: %s\nArguments: %s" % (self.step, self.stmt.__name__, self.args)

    def run(self):
        """Run statement and return output.

        Returns:
            str: Output from statement.
        """
        return self.stmt(**self.args)

    def parse(self):
        """Translate a step into a statement.

        Return:
            NoneType: If no statement was found, otherwise mixt.
        """
        for func in self.known_actions:
            for stmt in self.parse_func(func):
                result = re.match(stmt, self.step, re.I)
                if result is None:
                    continue
                return func, result.groupdict()
        return None

    def parse_func(self, func, statement_header=None, test=False):
        """Extract "accepted statements" from function doc string.

        Args:
            func (function): Statement function equivalent.
            statement_header (str): Headline to look up (default None).

        Raises:
            SystemExit: If `func` is not a function.

        Returns:
            list: List of possible statements.
        """
        if not callable(func):
            raise SystemExit("Expected callable function")
        if statement_header is None:
            statement_header = self.statement_header
        lines = func.__doc__.split("\n")
        return Parser.parse(tuple(lines), [statement_header], test=test)



# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import print_function

from re import compile
from importlib import import_module

from buildok.statement import Statement
from buildok.parser import Parser

class Tester(object):

    IMPORT_PATTERN = compile(r"^from\s+(?P<mod>buildok\.statements\.\w+)\s+import\s+(?P<stmt>\w+)$")
    STATEMENT_PY = "buildok/statement.py"
    STATEMENTS_MOD = "buildok.statements"
    TEST_INPUT_OUTPUT = ("sample input", "expected")

    @classmethod
    def scan(cls):
        imports, ctx = [], []
        with open(cls.STATEMENT_PY) as file_:
            ctx = file_.readlines()
        for line in ctx:
            if not line.startswith("from"):
                continue
            result = cls.IMPORT_PATTERN.match(line)
            if result is None:
                continue
            statement = result.group("stmt").strip()
            module = result.group("mod").strip()
            imports.append((module, statement))
        return imports

    @classmethod
    def launch_test(cls, module, func):
        tin, tout = cls.TEST_INPUT_OUTPUT
        mod = import_module(module)
        if not hasattr(mod, func):
            raise ImportError
        test = getattr(mod, func)
        test_doc = test.__doc__.split("\n")
        steps = Parser.parse(test_doc, [tin])
        if not Parser.all_valid(steps):
            raise ImportError
        expected = Parser.parse(test_doc, [tout])
        pr = Parser(tuple(steps))
        pr.prepare()
        return pr, test, expected

    @classmethod
    def run_test(cls, imports):
        module, func = imports
        output = "n/a"
        try:
            pr, test, expected = cls.launch_test(module, func)
            if len(expected) != 1:
                raise Exception("Incomplete or invalid test: no expectations")
            else:
                expected = expected[0]
            print("[Test] Running tests on %s" % module)
            while pr.has_step():
                step = pr.get_step()
                print("[Test]  ", step)
                stmt = Statement(step)
                output = stmt.run()
                if output is None:
                    raise Exception("Unexpected statement '%s', maybe misspeled" % step)
            assert output == expected, "Expected '%s' got '%s'" % (expected, output)
            print(u"[Test] \033[92mPassed OK %s\033[0m" % module)
        except ImportError:
            print(u"[Test] \033[93mMissing test: %s\033[0m" % module)
        except AssertionError as e:
            print(u"[Test] \033[101mFailed %s: %s\033[0m" % (module, e))
        except Exception as e:
            print(u"[Test] \033[91mError %s: %s\033[0m" % (module, e))
        print("[Test]", "-" * 50)

def main():
    for t in Tester.scan():
        try:
            Tester.run_test(t)
        except Exception as e:
            raise



# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import os

from converters.bash import unpack_bash
from util.console import Console


class Converter(object):
    """Translate build steps to script files.

    Used as a converter wrapper for multiple scriping files.

    Args:
        workdir (str): Working directory.
        target (tuple): Unpacked holder for script parameters.
        statements (frozenset): Set of all known statements.
    """
    workdir = os.getcwd()
    target, statements = None, None

    @classmethod
    def prepare(cls, target, statement):
        if target == "bash":
            cls.target = unpack_bash()
        elif target == "vagrant":
            raise Exception("Unsupported yet: %s" % target)
        elif target == "docker":
            raise Exception("Unsupported yet: %s" % target)
        elif target == "jenkins":
            raise Exception("Unsupported yet: %s" % target)
        elif target == "ansible":
            raise Exception("Unsupported yet: %s" % target)
        else:
            raise Exception("Unsupported target: %s" % target)
        cls.statements = statement.statements

    @classmethod
    def check(cls):
        return cls.target is not None and len(cls.statements) > 0

    @classmethod
    def save(cls, lines=[]):
        lang, fname, template = cls.target
        for s in cls.statements:
            func = lang.get(s.stmt.__name__)
            if not callable(func):
                continue
            lines.append(func(**s.args))
        with open(os.path.join(cls.workdir, fname), "w") as file_:
            data = template.format("\n".join(lines))
            file_.write(data)
        Console.info("Converted %d steps to %s file" % (len(lines), fname))



# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from buildok.reader import Reader
from buildok.parser import Parser

class ReadmeReader(Reader):
    """File class helper to detect and read a README file.

    Attributes:
        filename (string): Path to README filename.
        context (string): Original context of `filename`.
        build_section (tuple): Build section accepted titles.
    """
    READ_MODE = "rb"
    filename = "README.md"
    context = []
    build_section = ("how to build ok", "build ok steps")

    def read(self):
        """Read README file content.

        Returns:
            str: Returns build steps from `filename` content on success.
            NoneType: Returns None on failure.

        Raises:
            IOError: If `filename` does not exist or does not have permissions.
        """
        with open(self.filename, self.READ_MODE) as file_:
            ctx = file_.readlines()
            self.parse_build(ctx)
            return self.context
        return None

    def parse_build(self, ctx=[]):
        """Parse README file and look for build steps.

        Args:
            ctx (list): Original `filename` content as lines.

        Returns:
            list: Non-empty list if build steps are found.
        """
        self.context = Parser.parse(tuple(ctx), self.build_section)
